class ValidationWarning(UserWarning):
    """Warning for when validation fails."""

    pass


class AlignmentWarning(UserWarning):
    """Warning for when header and data are not aligned."""

    pass
